;
; Example program for pathfinding routine
;
;22/07/2001  Olivier Lecocq  olecocq@hotmail.com
;

#include "ti86asm.inc"
#include "asm86.h"
#include "ti86abs.inc"
 
_pause = $4d43
_ldhlz = $437B

map = $9000
pmap = map + $80        ; $80 = 128 = 8 * 16
mem1_add = $8100
Mem2_add = mem1_add + 160

startx = _textShadow
starty = startx+1
endx = starty+1
endy = endx+1
curx = endy+1
cury = curx+1
count = cury+1
cur_add = count+1       ;2 bytes
temp_add = cur_add+2    ;2 bytes
path_found = temp_add+2
tempx = path_found+1
tempy = tempx+1
dir = tempy+1
showflag = dir+1
path = showflag+1

.org _asm_exec_ram

    call _runindicoff
    call _clrLCD
    
Init_vars:
    ld a,0
    ld (showflag),a
    ld a,7
    ld (curx),a
    ld a,3
    ld (cury),a
    call Clear_map
    ld a,100
    ld hl,startx
    ld (hl),a
    inc hl
    ld (hl),a
    inc hl
    ld (hl),a
    inc hl
    ld (hl),a
    
Clear_all:    
    call Clear_pmap
    call Clear_map0
    call Update_display
         
Mainloop:
    halt
    call GET_KEY
    cp K_SECOND
    jp z,edit_map
    cp K_alpha
    jp z,set_start
    cp K_xvar
    jp z,set_goal
    cp K_ENTER
    jp z,Find_path
    cp K_CLEAR
    jp z,Clear_all
    cp K_GRAPH
    jp z,showonoff
    cp K_UP
    jp z,Move_up
    cp K_DOWN
    jp z,Move_down
    cp K_RIGHT
    jp z,Move_right
    cp K_LEFT
    jp z,Move_left
    cp K_exit
    jp z,Exit
    jp Mainloop    

Move_up:
    ld a,(cury)
    cp 0
    jp z,Mainloop
    call Cursor
    ld a,(cury)
    dec a
    ld (cury),a
    call Cursor
    jp Mainloop
    
Move_down:
    ld a,(cury)
    cp 7
    jp z,Mainloop
    call Cursor
    ld a,(cury)
    inc a
    ld (cury),a
    call Cursor
    jp Mainloop
                
Move_right:
    ld a,(curx)
    cp 15
    jp z,Mainloop
    call Cursor
    ld a,(curx)
    inc a
    ld (curx),a
    call Cursor
    jp Mainloop
    
Move_left:
    ld a,(curx)
    cp 0
    jp z,Mainloop
    call Cursor
    ld a,(curx)
    dec a
    ld (curx),a
    call Cursor
    jp Mainloop

set_start:
    ld hl,map
    ld b,128
start_loop:
    ld a,(hl)
    cp 5
    jp z,delete_start
    inc hl    
    djnz Start_loop
delete_start:
    ld (hl),0    
    ld hl,curx
    ld b,(hl)
    inc hl
    ld c,(hl)
    call Calc_map_pos
    ld (hl),5
    call Update_display
    ld a,(curx)
    ld hl,startx
    ld (hl),a
    ld a,(cury)
    ld hl,starty
    ld (hl),a
    jp Mainloop
    
set_goal:
    ld b,128
    ld hl,map
goal_loop:
    ld a,(hl)
    cp 6
    jp z,delete_goal
    inc hl
    djnz goal_loop
delete_goal:   
    ld (hl),0
    ld hl,curx
    ld b,(hl)
    inc hl
    ld c,(hl)
    call Calc_map_pos
    ld (hl),6
    call Update_display
    ld a,(curx)
    ld hl,endx
    ld (hl),a
    ld a,(cury)
    ld hl,endy
    ld (hl),a    
    jp Mainloop

Showonoff:
    ld a,(showflag)
    cpl
    ld (showflag),a
    jp Mainloop
    
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

; Breadth-first pathfinding algorithm

;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                
find_path:
    ld a,0
    ld (path_found),a
    ld a,(startx)    
    cp 100
    jp z,Mainloop
    ld a,(endx)
    cp 100
    jp z,Mainloop 
    call Clear_pmap
    call delete_mem1
    call delete_mem2
    ld a,1
    ld (count),a
    ld hl,mem1_add
    ld (temp_add),hl        ; save temp address for next open road               
    ld (hl),$ff             ; stopcode
    ld a,(startx)
    ld b,a
    ld a,(starty)
    ld c,a       
    call calc_map_pos
    ld de,$80
    add hl,de
    ld (hl),100             ; 100= code on pmap for starting point
    ld hl,mem2_add
    ld a,(startx)
    ld (hl),a
    ld a,(starty)
    inc hl
    ld (hl),a               ; save start coord
    inc hl        
    ld (hl),$ff             ; stopcode
    ld hl,mem2_add
    ld (cur_add),hl
    
Next_step:
    ld hl,(cur_add)
    ld a,(hl)
    cp $ff                  ;stopcode?
    jp z,step_done
    ld (tempx),a
    inc hl
    ld a,(hl)
    ld (tempy),a                  
    cp 0                    ; check if upper border
    jp z, skip_up           ; upper border reached    
    ld c,a
    ld a,(tempx)
    ld b,a
    dec c                   ; check up 
    call Check_dir    
skip_up:    
    ld a,(tempx)
    cp 15
    jp z, skip_right        ; right border reached
    ld b,a
    inc b                   ; check right
    ld a,(tempy)
    ld c,a
    call Check_dir
skip_right:
    ld a,(tempy)
    cp 7
    jp z,Skip_down          ; lower border reached
    ld c,a
    inc c                   ; check lower
    ld a,(tempx)
    ld b,a
    call Check_dir
skip_down:
    ld a,(tempx)
    cp 0
    jp z,skip_done          ; left border reached
    ld b,a
    dec b                   ; check left
    ld a,(tempy)
    ld c,a
    call Check_dir
skip_done:
    ld a,(path_found)
    cp 1
    jp z,calc_path           ; path is found  
    ld hl,(cur_add)
    inc hl
    inc hl
    ld (cur_add),hl
    jp Next_step
Step_done:    
    ld a,(count)
    cp 80                   ; max steps
    jp z,Mainloop           
    inc a
    ld (count),a
    rra
    jp c,odd_count
    jp even_count
                 
Even_count:
    call delete_mem2                  
    ld hl,mem1_add
    ld (cur_add),hl
    ld hl,mem2_add
    ld (temp_add),hl
    ld (hl),$FF             ; stopcode
    jp next_step
Odd_count:
    call delete_mem1
    ld hl,mem2_add
    ld (cur_add),hl
    ld hl,mem1_add
    ld (temp_add),hl 
    ld (hl),$ff             ;stopcode
    jp next_step

Check_dir:
    push bc
    call Calc_map_pos
    pop bc
    ld de,$80
    add hl,de               ; pmap address
    ld a,(hl)               ; get pmap byte
    cp 0
    ret nz                  ; tile already checked    
    sbc hl,de               ; map address
    ld a,(hl)               ; get map byte
    cp 0
    call z,Add_free_node
    cp 6                    ; 6=goal
    call z,goal_reached
    ret

Add_free_node:
    ld de,$80
    add hl,de
    ld a,(count)
    ld (hl),a               ; save distance on pmap
    ld hl,(temp_add)
    ld (hl),b
    inc hl
    ld (hl),c               ; save coord
    inc hl
    ld (hl),$FF             ; stopcode
    ld (temp_add),hl        ; save next mem slot 
    ld a,(showflag)
    cp 255
    call z,display_progress
    xor a
    ret
    
Display_progress:                
    call Calc_xy  
    ld hl,$c37c
    ld (hl),b
    inc hl
    ld (hl),c
    ld a,(count)
    ld h,0
    ld l,a  
    call DisplayHL
    ld b,20
    call delay
    ret   

Calc_path:
    ld a,(count)
    ld hl,path
    ld b,0
    ld c,a
    add hl,bc
    ld (cur_add),hl
    ld a,(endx)
    ld (tempx),a
    ld a,(endy)
    ld (tempy),a
    ld a,(count)
    cp 1
    jp nz,Calc_path_loop
    ;ld a,101
    ld a,100
    ld (count),a
Calc_path_loop:    
    ld a,3
    ld (dir),a          
    ld a,(tempx)
    ld b,a
    ld a,(tempy)
    ld c,a    
    ld a,c
    cp 0
    jp z,skip_dir1
    dec a
    ld c,a
    call check_tile
skip_dir1:
    ld a,4
    ld (dir),a
    ld a,(tempx)
    ld b,a
    ld a,(tempy)
    ld c,a
    ld a,b
    cp 15
    jp z,skip_dir2
    inc a
    ld b,a
    call check_tile
skip_dir2:
    ld a,1
    ld (dir),a
    ld a,(tempx)
    ld b,a
    ld a,(tempy)
    ld c,a
    ld a,c
    cp 7
    jp z,skip_dir3
    inc a
    ld c,a
    call check_tile
skip_dir3:
    ld a,2
    ld (dir),a
    ld a,(tempx)
    ld b,a
    ld a,(tempy)
    ld c,a
    ld a,b
    cp 0
    jp z,skip_dir4
    dec a
    ld b,a
    call check_tile
skip_dir4:
    ld a,(count)
    cp 100
    jp nz,calc_path_loop
    call update_display
    jp Mainloop                     ; path found    

Check_tile:
    push bc    
    call calc_map_pos
    pop bc
    ld de,$80
    add hl,de
    ld a,(count)
    dec a
    cp (hl)
    ret nz
    ld a,(dir)
    ld hl,(cur_add)
    ld (hl),a
    dec hl
    ld (cur_add),hl
    ld a,b
    ld (tempx),a
    ld a,c
    ld (tempy),a
    call calc_map_pos
    ld a,(dir)
    ld (hl),a 
    ld a,(count)
    cp 2
    jp z,almost_there
    dec a
    ld (count),a
    ret
almost_there:
    ;ld a,101
    ld a,100
    ld (count),a
    ret

                            
Goal_reached:
    ld a,1
    ld (path_found),a
    ret   
                
delete_mem1:
    ld hl,mem1_add
    ld b,160
    call _ldhlz
    ret
    
delete_mem2:
    ld hl,mem2_add
    ld b,160
    call _ldhlz
    ret

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                                                                               
Cursor:
    ld a,(curx)
    ld c,a
    ld a,(cury)
    ld b,a
    call Calc_screen_pos
    ld b,8
    ld de,16
Cursor_loop:    
    ld a,(hl)
    cpl
    ld (hl),a
    add hl,de
    djnz Cursor_loop
    ret

edit_map:
    ld hl,curx
    ld b,(hl)
    inc hl
    ld c,(hl)
    call calc_map_pos
    ld a,(hl)
    cp 0
    jp nz,not_empty
    ld (hl),7              ; 7=block    
    call Update_display
    jp Mainloop    
not_empty:
    cp 7
    jp nz,Mainloop
    ld (hl),0
    call Update_display
    jp Mainloop

delay:    
    halt
    djnz delay
    ret
        
Exit:
    ret    
        
Clear_map:
    ld hl,map
    ld b,$80
    call _ldhlz
    ret

Clear_map0:
    ld hl,map
    ld b,128
Clear_map0_loop:
    ld a,(hl)
    cp 5
    jp p,not_delete
    ld (hl),0    
not_delete:
    inc hl
    djnz Clear_map0_loop
    ret
                                                                
Clear_pmap:
    ld hl,pmap
    ld b,$80
    call _ldhlz
    ret

Update_display:
    call Drawmap
    call Cursor
    ret

DrawMap:
 ld de,$0000            ; start drawing at (0, 0)
 ld hl,map
DrawMapL:
 push hl                ; save map pointer
 ld l,(hl)              ; get the current tile
 ld h,0                 ; clear the upper byte
 add hl,hl              ; multiply by 8 (* 2)
 add hl,hl              ; * 4
 add hl,hl              ; * 8
 ld bc,Sprites          ; load the start of the sprites
 add hl,bc              ; add togehter to get correct sprite
 push de                ; save sprite draw location
 call GridPutSprite     ; draw the sprite
 pop de                 ; restore sprite draw location
 pop hl                 ; restore map pointer
 inc hl                 ; go to next tile
 inc e                  ; move location to the right
 bit 4,e                ; if the 5th bit (e = 16), we're done
 jr z,DrawMapL          ; only jump if the row isn't complete
 ld e,0                 ; draw at the left again
 inc d                  ; move location down a row
 bit 3,d                ; check if it's done (e = 8)
 ret nz                 ; return if we're done
 jr DrawMapL            ; jump back to the top of the loop

;===========================================================
; GridPutSprite:                       [routine by Dan Eble]
;  Puts an aligned sprite at (E, D), where HL -> Sprite
;===========================================================
; IN : D  = y (0-7 inclusive)
;      E  = x (0-15 inclusive)
;      HL-> sprite
;
; OUT: AF, BC, DE, HL, IX modified
; Current total : 28b/567t (not counting ret)
;===========================================================
GridPutSprite:
 push hl
 pop ix                 ; ix-> sprite
 srl d                  ; de = 128y+x (16 bytes/row, 8 rows)
 rra
 and $80
 or e                   ; add x offset (remember x <= 15)
 ld e,a
 ld hl,$fc00            ; hl-> video ram
 add hl,de              ; hl-> video ram + offset
 ld b,8                 ; initialize loop counter
 ld de,$10              ; initialize line increment
GPS_Loop
 ld a,(ix+0)            ; get byte from sprite
 ld (hl),a              ; put byte on screen
 inc ix                 ; move to next byte in sprite
 add hl,de              ; move to next line on screen
 djnz GPS_Loop
 ret        

Calc_map_pos:           ; input b=x coord (0-15); c=y coord (0-7); output hl=map address  
    ld d,b
    ld a,c
    add a,a     ;x2
    add a,a     ;x4
    add a,a     ;x8
    add a,a     ;x16
    ld hl,map
    ld b,0
    ld c,a
    add hl,bc
    ld c,d
    add hl,bc
    ret
    
Calc_screen_pos:        ; input c=x coord (0-15); b=y coord (0-7); output hl=display address 
    ld hl,$fc00
    ld a,b
    cp 0
    jr z,Get_start_screen
    ld de,128
Next_line_screen:
    add hl,de
    djnz Next_line_screen
Get_start_screen:
    add hl,bc
    ret    

Calc_xy:                ; input b=x coord (0-15);c=y coord (0-7); output b=x coord (0-127);c=y coord (0-63)
    ld a,b
    add a,a
    add a,a
    add a,a
    ld b,a
    ld a,c
    add a,a
    add a,a
    add a,a
    ld c,a
    ret

DisplayHL:
    ld de,_OP1+5     
    ld b,2          
    call UnpDigits2 
    ex de,hl        
    call _vputs
    ret
UnpDigits2: 
    xor a           
    ld (de),a       
Unpack: 
    call UNPACK_HL  
    add a,'0'       
    dec de          
    ld (de),a      
    djnz Unpack
    ret
    
                                      
Sprites:
    .db %00000000
    .db %00000000
    .db %00000000
    .db %00000000
    .db %00000000
    .db %00000000
    .db %00000000
    .db %00000000
                
Up_arrow:
    .db %00011000
    .db %00100100
    .db %01000010
    .db %00100100
    .db %00100100
    .db %00100100
    .db %00111100
    .db %00000000

Right_arrow:
    .db %00000000
    .db %00000100
    .db %01111010
    .db %01000001
    .db %01000001
    .db %01111010
    .db %00000100
    .db %00000000

Down_arrow:    
    .db %00000000
    .db %00111100
    .db %00100100
    .db %00100100
    .db %00100100
    .db %01000010
    .db %00100100
    .db %00011000
    
Left_arrow:    
    .db %00000000
    .db %00100000
    .db %01011110
    .db %10000010
    .db %10000010
    .db %01011110
    .db %00100000
    .db %00000000

Start_sprite:    
    .db %00111110
    .db %01000000
    .db %01000000
    .db %00111100
    .db %00000010
    .db %00000010
    .db %01111100
    .db %00000000

Goal_sprite:    
    .db %00111100
    .db %01000010
    .db %01000000
    .db %01001110
    .db %01000010
    .db %01000010
    .db %00111100
    .db %00000000

Block_sprite:    
    .db %11111111
    .db %11111111
    .db %11111111
    .db %11111111
    .db %11111111
    .db %11111111
    .db %11111111
    .db %11111111
   
.end    